import pytest
from fantasy_ga.configs import ContestConfig, ModelConfig, Site, League


# Normal cases for ContestConfig
def test_contest_config_default_values():
    config = ContestConfig()
    assert config.site == Site.DK
    assert config.league == League.NBA


def test_contest_config_custom_values():
    config = ContestConfig(site=Site.FD, league=League.NFL)
    assert config.site == Site.FD
    assert config.league == League.NFL


# Failure cases for ContestConfig
def test_contest_config_invalid_site():
    with pytest.raises(ValueError):
        ContestConfig(site="invalid_site", league=League.NBA)


def test_contest_config_invalid_league():
    with pytest.raises(ValueError):
        ContestConfig(site=Site.DK, league="invalid_league")


# Normal cases for ModelConfig
def test_model_config_default_values():
    config = ModelConfig()
    assert config.n_pop == 1000
    assert config.n_breed == 30
    assert config.n_mutate == 30
    assert config.n_gen == 16
    assert config.n_compound == 5


def test_model_config_custom_values():
    config = ModelConfig(n_pop=500, n_breed=20, n_mutate=20, n_gen=8, n_compound=3)
    assert config.n_pop == 500
    assert config.n_breed == 20
    assert config.n_mutate == 20
    assert config.n_gen == 8
    assert config.n_compound == 3


# Failure cases for ModelConfig
def test_model_config_negative_values():
    with pytest.raises(ValueError):
        ModelConfig(n_pop=-1, n_breed=-1, n_mutate=-1, n_gen=-1, n_compound=-1)


def test_model_config_non_integer_values():
    with pytest.raises(ValueError):
        ModelConfig(
            n_pop=1,
            n_breed=[1],
            n_mutate=3,
            n_gen=4,
            n_compound=5,
        )



import pytest
import os
from fantasy_ga.lineup_generator import LineupGenerator
from fantasy_ga.configs import Site, League, ModelConfig, ContestConfig

@pytest.fixture
def sample_data_path(tmpdir):
    return "examples/DraftKings/NBA/DKSalaries.csv"

@pytest.fixture
def export_path(tmpdir):
    return str(tmpdir.join("output.csv"))

def test_lineup_generator(sample_data_path, export_path):
    args = {
        "data_path": sample_data_path,
        "export_path": export_path,
        "site": "DK",
        "league": "NBA",
        "top_n_lineups": "1",
        "n_pop": "100",
        "n_gen": "20",
        "n_breed": "10",
        "n_mutate": "5",
        "n_compound": "5",
    }

    cc = ContestConfig(site=Site.DK, league=League.NBA)
    mc = ModelConfig(
        n_pop=int(args["n_pop"]),
        n_gen=int(args["n_gen"]),
        n_breed=int(args["n_breed"]),
        n_mutate=int(args["n_mutate"]),
        n_compound=int(args["n_compound"]),
    )

    model = LineupGenerator(cc, mc)
    model.read_csv(args["data_path"])
    model.fit()
    model.export_csv(args["export_path"], top_n=int(args["top_n_lineups"]))

    assert os.path.exists(args["data_path"])
    assert os.path.exists(args["export_path"])

    lineups, scores = model.get_top_n_lineups(1)
    assert len(lineups) == 1
    assert scores[0] >= 0




import numpy as np
from fantasy_ga import LineupGenerator
from fantasy_ga.configs import ModelConfig, ContestConfig, Site, League


def test_positions_random_lineups():
    cc = ContestConfig(site=Site.DK, league=League.NBA)
    mc = ModelConfig()

    model = LineupGenerator(cc, mc)
    model.set_matrix(np.loadtxt("tests/test_data.csv", delimiter=",", skiprows=1))
    lineups = model.create_random_lineups()

    for l in lineups:
        player_ids = [int(id) for id in l]

        # test there's no overlap
        assert len(player_ids) == len(set(player_ids))

        # test they're all eligible for their position
        assert np.diag(model.m.take(player_ids, 0)[:, 3:]).sum() == len(
            model._positions
        )



"""
Configurations for Contests and Optimizer Model
"""

from enum import Enum
from typing import Optional
from pydantic import BaseModel, conint


class League(Enum):
    """
    Enum for different sports leagues. Common across different platforms.

    Values
    - UNKNOWN: Unknown league
    - NBA: National Basketball Association
    - NFL: National Football League
    - MLB: Major League Baseball
    - NHL: National Hockey League
    """

    UNKNOWN = 0
    NBA = 1
    NFL = 2
    MLB = 3
    NHL = 4


class Site(Enum):
    """
    Enum for different fantasy sports platforms.

    Values
    - UNKNOWN: Unknown platform
    - DK: DraftKings
    - FD: FanDuel
    """

    UNKNOWN = 0
    DK = 1
    FD = 2


POSITIONS = {
    # https://www.draftkings.co.uk/help/rules/overview
    Site.DK: {
        League.NBA: ["PG", "SG", "SF", "PF", "C", "G", "F", "UTIL"],
        League.MLB: ["P", "P", "C", "1B", "2B", "3B", "SS", "OF", "OF", "OF"],
        League.NFL: ["QB", "RB", "RB", "WR", "WR", "WR", "TE", "FLEX", "DST"],
        League.NHL: ["C", "C", "W", "W", "W", "D", "D", "G", "UTIL"],
    }
}

SALARY_CAPS = {
    # https://www.draftkings.co.uk/help/rules/overview
    Site.DK: {
        League.NBA: 50000,
        League.MLB: 50000,
        League.NFL: 50000,
        League.NHL: 50000,
    }
}


class ContestConfig(BaseModel):
    """
    Config for a contest, specifying the site and league.

    Attributes:
        site (Site): The site where the contest is hosted. Currently only DraftKings is supported.
        league (League): The sports league associated with the contest.
    """

    site: Site = Site.DK
    league: League = League.NBA


class ModelConfig(BaseModel):
    """
    Config for the optimizer, specifying parameters for the genetic algorithm.

    Attributes:
        n_pop (Optional[conint(ge=0)]): Size of the initial population.
        n_breed (Optional[conint(ge=0)]): Number of player pairs to breed in each generation.
        n_mutate (Optional[conint(ge=0)]): Number to players to mutate in each generation.
        n_gen (Optional[conint(ge=0)]): Number of generations.
        n_compound (Optional[conint(ge=0)]): Size of initial populations to compound over.

    Note:
        All parameters are optional and have default values if not specified.
    """

    n_pop: Optional[conint(ge=0)] = 1000
    n_breed: Optional[conint(ge=0)] = 30
    n_mutate: Optional[conint(ge=0)] = 30
    n_gen: Optional[conint(ge=0)] = 16
    n_compound: Optional[conint(ge=0)] = 5



"""
LineupGenerator to read csv files, optimize the player lineup, and return the list of lineups.
"""

import csv
from typing import List
import numpy as np
from fantasy_ga.configs import (
    Site,
    ContestConfig,
    ModelConfig,
    POSITIONS,
    SALARY_CAPS,
)


class LineupGenerator:
    def __init__(
        self,
        cc: ContestConfig,
        mc: ModelConfig,
    ):
        """
        Initialize the LineupGenerator.

        Args:
            cc (ContestConfig): Contest configuration including site and league information.
            mc (ModelConfig): Model configuration including parameters for the genetic algorithm.
        """
        self._cc: ContestConfig = cc
        self._mc: ModelConfig = mc
        self._positions: List[str] = POSITIONS[self._cc.site][self._cc.league]
        self._salary_cap: int = SALARY_CAPS[self._cc.site][self._cc.league]
        self.id_to_name: dict = {}
        self.id_to_salary: dict = {}
        self.m: np.ndarray = None
        self.lineups: np.ndarray = []
        self.scores: np.ndarray = []
        self.pos_start_idx: int = 3

    def encode_position(self, s: str) -> list[int]:
        """
        Encode a position name in a string format into a binary vector representation.
        For example, PG/SG (Point Guard/Shooting Guard) is [1,1,0,0,0,0,0,1] where the first two 1s represent
        PG and SG and the last one UTIL (Utility) which is 1 for any position.

        Args:
            s (str): position name

        Returns:
            list[int]: a binary vector of positions where 1 represents this position being applicable
        """
        res = [0] * len(self._positions)
        for pos in s.split("/"):
            for i in range(len(self._positions)):
                if pos == self._positions[i]:
                    res[i] = 1
                    # break if no duplicate positions allowed in the lineup
                    if len(self._positions) == len(set(self._positions)):
                        break
        return res

    def read_csv(self, filepath: str):
        """
        Read a CSV file from contest sites and constructs dictionary for player name, id, and salary
        and sets the player matrix consisting of name, id, salary and the position vector.

        Args:
            filepath (str): Path to the CSV file

        Raises:
            NotImplementedError: Raised for FanDuel format for now.
        """
        with open(filepath, mode="r") as f:
            reader = csv.reader(f)
            _ = next(reader)
            players = []
            if self._cc.site == Site.DK:
                for row in reader:
                    self.id_to_name[int(row[3])] = row[2]
                    self.id_to_salary[int(row[3])] = int(row[5])
                    players.append(
                        [int(row[3]), int((row[5])), float(row[-1])]
                        + self.encode_position(row[4])
                    )
            elif self._cc.site == Site.FD:
                raise NotImplementedError
        self.m = np.array(players)

    def set_matrix(self, m: np.array):
        """
        Set the player matrix directly using numpy array.

        Args:
            m (np.array): Player matrix where each row consists of id, salary, fpts followed by a vector of positions for the corresponding sport.
        """
        self.m = m

    def export_csv(self, filepath: str, top_n: int = None):
        """
        Export the lineups in a CSV format submittable in the contest.

        Args:
            filepath (str): Path to the output CSV file
            top_n (int, optional): Top N lineups to output. Defaults to None.
        """
        n = top_n if top_n else min(500, len(self.lineups))
        lineups, _ = self.get_top_n_lineups(n)

        with open(filepath, "w") as f:
            w = csv.writer(f)
            w.writerow(self._positions)
            for lineup in lineups:
                w.writerow([_id for _id in lineup.astype(int)])

    def get_top_n_lineups(self, n: int) -> np.ndarray:
        """
        Get the top N lineups based on predicted Fantasy Point totals.

        Args:
            n (int): Number of lineups to return

        Returns:
            np.ndarray: An array of lineups sorted by highest to lowest FPTS scores
        """
        top_n_scores = (-self.scores).argsort()[:n]
        return (
            self.lineups.take(top_n_scores.astype(int), 0),
            self.scores.take(top_n_scores.astype(int), 0)
        )

    def calc_scores(self):
        """
        Calculate scores for each lineup based on salary cap and position constraints.
        """
        scores = []
        for lineup in self.lineups:
            sal, fpts = self.m[np.in1d(self.m[:, 0], lineup.astype(int))][
                :, [1, 2]
            ].sum(axis=0)
            scores.append(fpts) if sal <= self._salary_cap and len(
                np.unique(lineup)
            ) == len(self._positions) else scores.append(-1)
        self.scores = np.array(scores)

    def create_random_lineups(self) -> np.ndarray:
        """
        Create random lineups.

        Returns:
            np.array: An array of random lineups satisfying the position constraints.
        """
        lineups = []

        for _ in range(self._mc.n_pop):
            lineup = []
            for pos_key in range(self.pos_start_idx, self.m.shape[1]):
                candidate_ids = self.m[self.m[:, pos_key] == 1][:, 0]
                searching = True
                while searching:
                    candidate = np.random.choice(candidate_ids)
                    if candidate not in lineup:
                        lineup.append(candidate)
                        searching = False
            lineups.append(lineup)
        return np.array(lineups)

    def breed(self):
        """
        Create new lineups by randomising the elements from the top two lineups in terms of scores.
        """
        new_lineups = []
        parents_idx = (-self.scores).argsort()[:2]
        parents = self.lineups.take(parents_idx, 0)

        for _ in range(self._mc.n_breed):
            if np.array_equal(*parents):
                self.lineups = parents
                return
            else:
                breed_idx = np.random.choice(2, len(self._positions))
                new_lineups.append([parents[p, idx] for idx, p in enumerate(breed_idx)])
        self.lineups = np.vstack([parents, np.array(new_lineups)])

    def mutate(self):
        """
        Create new lineups by randomly swapping elements
        """
        mutate_idx = np.random.choice(self.lineups.shape[0], self._mc.n_mutate)

        for idx in mutate_idx:
            searching = True
            while searching:
                # Search for two rows where positions are swappable
                mutant = self.m[np.random.choice(self.m.shape[0]), :]
                original = self.m[
                    self.m[:, 0] == np.random.choice(self.lineups[idx]).astype(int), :
                ][0]
                eligible_pos = np.where(
                    mutant[self.pos_start_idx :].astype(bool)
                    & original[self.pos_start_idx :].astype(bool)
                )[0]
                if len(eligible_pos) > 0:
                    searching = False
            swap_pos = np.random.choice(eligible_pos)
            self.lineups = np.vstack([self.lineups, self.lineups[idx]])
            self.lineups[-1, swap_pos] = mutant[0]

    def evolve(self):
        """
        Repeat the breed and mutate processes
        """
        for _ in range(self._mc.n_gen):
            self.breed()
            self.mutate()
            self.calc_scores()

    def fit(self):
        """
        Repeat the evolve process with several initial populations
        """
        self.lineups = self.create_random_lineups()
        for _ in range(self._mc.n_compound):
            self.calc_scores()
            self.evolve()
            self.lineups = np.vstack([self.lineups, self.create_random_lineups()])



import argparse
from fantasy_ga.lineup_generator import LineupGenerator
from fantasy_ga.configs import Site, League, ModelConfig, ContestConfig


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_path",
        help="File path to the csv file exported from daily fantasy platforms",
    )
    parser.add_argument(
        "--export_path",
        help="File path to export the top lineups",
    )
    parser.add_argument("--site", help="Site name. Currently supports DraftKings/DK")
    parser.add_argument("--league", help="League name. Currently supports NBA")
    parser.add_argument("--top_n_lineups", help="top N lineups to display")
    parser.add_argument("--n_pop", help="Number of initial population to sample from")
    parser.add_argument("--n_gen", help="Number of generations to run evolution cycles")
    parser.add_argument("--n_breed", help="Number of children lineups to generate")
    parser.add_argument("--n_mutate", help="Number of mutations for lineups")
    parser.add_argument("--n_compound", help="Number of compound evolution cycles")

    args = parser.parse_args()

    cc = ContestConfig(
        Site.DK if args.site.lower() in ["dk", "draftkings"] else Site.UNKNOWN,
        League.NBA if args.league.lower() == "nba" else League.UNKNOWN,
    )
    mc = ModelConfig(
        n_pop=int(args.n_pop),
        n_gen=int(args.n_gen),
        n_breed=int(args.n_breed),
        n_mutate=int(args.n_mutate),
        n_compound=int(args.n_compound),
    )

    model = LineupGenerator(cc, mc)
    model.read_csv(args.data_path)
    model.fit()
    model.export_csv(args.export_path, top_n=int(args.top_n_lineups))
    print(f"Top {args.top_n_lineups} lineups exported into {args.export_path}")

    lineups, scores = model.get_top_n_lineups(1)
    print(
        f"[Best Lineup]\nPlayers: {[model.id_to_name[id] for id in lineups[0]]}\nSalary Total: {sum([model.id_to_salary[id] for id in lineups[0]])}\nExpected FPTS: {scores[0]}"
    )


if __name__ == "__main__":
    main()
